#!/usr/bin/env python

# analysis process
# Manuel Amador (Rudd-O)
# part of Smart DJ
# written on 2005-11-28

from songanalysiscommon import *
import logging
from ControllableThread import ControllableThread
from Observable import Observable,Observer
from amarokplus import NotAnalyzedYet, UnAnalyzable, DcopAmarokFailed
from songanalysiswrapper import analyze_song, get_interface, NotSong
from AsyncProcess import AsyncProcess,NotReadyYet
from Queue import Empty
from songanalysiscommon import say_sorry,n_logfile
import time
from extattr import getfattr,setfattr,delfattr
import errno


class AnalysisProcess(ControllableThread,Observable,Observer):
	
	logger = logging.getLogger("AnalysisProcess")
	
	def __init__(self,proxyObject):
		ControllableThread.__init__(self)
		Observable.__init__(self)
		Observer.__init__(self)
		self.analysis_priority_queue = []
		self.amarok = proxyObject
		self.observe(self.amarok)
		self.observe(self)
		self.amarok_available_yet = False
		
	def addToPriorityQueue(self,filename): # we need to serialize this around a common lock X
		if not filename in self.analysis_priority_queue:
			self.logger.debug("Adding to priority queue: %s",filename)
			self.analysis_priority_queue.append(filename)
			self.broadcastEvent("priorityQueueChanged")
		
	def removeFromPriorityQueue(self,filename):
		if filename in self.analysis_priority_queue:
			self.logger.debug("Removing from priority queue: %s",filename)
			self.analysis_priority_queue.remove(filename)
		
	def getPriorityQueueLength(self):
		return len(self.analysis_priority_queue)
		
	def getNextSong(self): # we need to serialize this around a common lock X
		try: return self.analysis_priority_queue[0]
		except IndexError: return self.amarok.collection.getNextUnanalyzedSong()
	
	def run(self):
		try: self._run()
		except DcopAmarokFailed,e:
			self.logger.debug("Aborting because amaroK is unexpectedly gone: %s",e.output)
		except:
			self.logger.exception("Something wrong in background song analyzer, thread stopped")
			say_sorry("Something went wrong in the analyzer thread loop.  This is not your fault, but a bug in the Smart DJ plug-in.  Please report the contents of the file %s to the Smart DJ plug-in developers." %os.path.expanduser(n_logfile))
		
	def notify(self,object,eventName,*args):
		if eventName == "amaroKAvailable": self.amarok_available_yet = True
		if eventName == "amaroKExit": self.kill(asynchronous=True)
		if eventName == "priorityQueueChanged": self.runNow = True
	
	runNow = False
	
	def _run(self):
		
		self.logger.debug("Analyzer starting")
		
		analyzing_popup_shown = False
		done_popup_shown = False
		
		while not self.amarok_available_yet and not self.killed(): time.sleep(1)
		self.logger.debug("Analyzer started")
		
		while not self.killed(): #with this, we make this thread finish
			self.logger.debug("Doing one iteration now")
			next_song = self.getNextSong()
			if next_song:
				self.runNow = False
				if analyzing_popup_shown is False:
					self.amarok.playlist.shortStatusMessage("Smart DJ is analyzing your collection")
					analyzing_popup_shown = True
					done_popup_shown = False # we want to show this once a change in your collection is found
				self.processSong(next_song)
			else:
				if done_popup_shown is False:
					self.amarok.playlist.shortStatusMessage("Your music collection has been fully analyzed")
					done_popup_shown = True
					analyzing_popup_shown = False # the same goes for this
				self.logger.debug("sleeping")
				for seconds in range(30):
					if not self.killed() and not self.runNow: time.sleep(1)
				self.runNow = False
					
		
		self.logger.debug("Analyzer done")
			#we should rather simply sleep until the main thread signals that a rescan of the collection has just finished
			
	def processSong(self,file):
		self.logger.info("Processing: " + file)
		
		# we begin by assuming we're analyzing the file and using eas
		onlyCollect = False
		useExtattrs = True
		
		#initialize vars we'll use
		tempo = None
		freqbands = None
		voldiff = None
		
		songanalysis_interface_installed = get_interface()
		# make something that if this value changes from function call N to function call M, the purge_songs_not_analyzed_with_interface function is called FIXME
		try: # Let's see if the file has songanalysis EAs
			date_analyzed  = getfattr(file,n_date_analyzed)
			
			#we'll also check if the EAs in the file were analyzed by the same songanalysis interface as the one we have installed
			try: songanalysis_interface_used  = getfattr(file,n_songanalysis_interface)
			except KeyError: songanalysis_interface_used = None
			self.logger.debug("Songanalysis interface available: %s     used for this song: %s"%(songanalysis_interface_installed,songanalysis_interface_used))
			
			if songanalysis_interface_used == songanalysis_interface_installed:
				# if the interfaces match, it's okay to collect from EAs
				self.logger.debug("A-OK, we only collect")
				onlyCollect = True # since the interfaces match and there is a date of analysis we will only collect
			else:
				# interface mismatch, a reanalysis is in order
				self.logger.debug("Interface mismatch, we will reanalyze")
				
		except KeyError: # File did not have an EA
			self.logger.debug("No last analysis date EA found, assuming no EA thus need to reanalyze and restore in the EA")
			
		except IOError, e:
				if e.errno == errno.EOPNOTSUPP: #No support for EAs on the filesystem where the file is installed
					onlyCollect = False # we will analyze the song
					useExtattrs = False # we won't store it on eas
				elif e.errno == errno.ENOENT:
					onlyCollect = False
					useExtattrs = False
					self.logger.debug("The song %s does not exist, letting the songanalysis process catch this so this file is marked as invalid"%file)
				else: raise
		
		if useExtattrs and onlyCollect: #if we're using already available eas
			self.logger.debug( "collecting info from extattrs")
			# we collect the info from the eas
			try: tempo = float(getfattr(file,n_tempo))
			except KeyError: pass
			try:
				freqbands = getfattr(file,n_freqbands)
				freqbands = freqbands.split(",")
				freqbands = [ float(f) for f in freqbands ]
			except KeyError: pass
			try: voldiff = float(getfattr(file,n_voldiff))
			except KeyError: pass
			self.logger.debug( "collected %s %s %s"%(tempo,freqbands,voldiff))
		else:
			process = AsyncProcess(analyze_song,(file,))
			process.start()
			
			# and periodically check whether it is done
			while not self.killed(): # we periodically check whether we should abandon this
				try:
					tempo,freqbands,voldiff=process.get(block=True,timeout=1)
					break
				except NotReadyYet:
					continue
				except NotSong:
					# this is not a song, but we will mark it as NULL NULL NULL
					self.logger.debug( "Not a song!")
					break
		
			if self.killed():
				self.logger.debug("shutdown requested, abandoning ongoing analysis")
				return
			
			if tempo != None and freqbands != None and voldiff != None:
				self.logger.debug( "deduced %s %s %s"%(tempo,freqbands,voldiff))
			
			if useExtattrs: # we store the info in eas
				try:
					for a in [ n_tempo,n_freqbands,n_voldiff,n_date_analyzed,n_songanalysis_interface]:
						try: delfattr(file,a) # let's wipe the attributes first
						except KeyError,e: pass # if there wasn't an attribute, go go gadget, ignore the evil!
					if tempo is not None: setfattr(file,n_tempo,str(tempo))
					if freqbands is not None: setfattr(file,n_freqbands,",".join([str(f) for f in freqbands]))
					if voldiff is not None: setfattr(file,n_voldiff,str(voldiff))
					date_analyzed = os.stat(file).st_mtime
					setfattr(file,n_date_analyzed,str(date_analyzed))
					setfattr(file,n_songanalysis_interface,songanalysis_interface_installed)
				except IOError,e:
					if e.errno == errno.EACCES: self.logger.debug( "no permission to save EAs for file %s, ignoring and moving on"%(file))
					elif e.errno == errno.ENOENT: self.logger.warning( "file %s vanished during analysis, saving in database, ignoring EA saves and moving on"%(file))
					else: raise
			
		# finally, we store the info in the DB
		self.amarok.collection.associate(file,tempo,freqbands,voldiff,songanalysis_interface_installed)
		# and remove it from the priority queue
		self.removeFromPriorityQueue(file)
		self.broadcastEvent("trackAnalyzed",file)
